// Copyright 2016, University of Colorado Boulder


(function() {
  'use strict';

  /**
   * Used to record a blank data point and give it to the callback
   * @param simIFrameClient - deprecated and not used in classroom activity
   * @callback callback
   */
  function recordBlankDataPoint( simIFrameClient, callback ) {
    var dataPoint = {};
    callback( dataPoint );
  }

  /**
   * Restore a given case to the sim
   * @param caseNumber
   * @param callback
   */
  function restoreCase( caseNumber, callback ) {
    window.simIFrameClient.invoke( window.simConfiguration.modelTandem, 'reset', [], function() {
      window.simIFrameClient.invokeSequence(
        window.simConfiguration.customization.applyCases[ caseNumber - 1 ]
          .concat( window.simConfiguration.customization.applyRestoreCase ), callback
      );
    } );
  }

  /**
   * validates if all of the cases have been filled out completely before drawing the data table
   * @returns {boolean} - if the apply tab is validated
   */
  function validateFillOutCases() {
    var $tabs = $( '.case.tab' );
    var flagCounter = 0;
    $tabs.each( function( i ) {
      var flag = false;
      $( 'input[name=apply-case' + (i + 1) + ']' ).each( function() {
        if ( !flag && this.checked ) {
          flag = true;
        }
      } );

      flag = flag && !!$( '#apply-case-textarea' + ( i + 1 ) ).val();

      if ( flag ) {
        if ( $tabs[ i ].classList.contains( 'incomplete' ) ) {
          $tabs[ i ].classList.remove( 'incomplete' );
          $tabs[ i ].classList.add( 'pickable' );
        }
        flagCounter++;
        window.classroom.apply.readyToTestCases = true;
      }
      else {
        if ( $tabs[ i ].classList.contains( 'pickable' ) ) {
          $tabs[ i ].classList.remove( 'pickable' );
          $tabs[ i ].classList.add( 'incomplete' );
        }
        window.classroom.apply.readyToTestCases = false;
      }
    } );

    var submitPredictionsButton = document.getElementById( 'apply-predictions-button' );

    var numCases = window.simConfiguration.customization.applyCases.length;

    if ( flagCounter >= numCases ) {
      // Enable/Disable the submit button if needed
      submitPredictionsButton.classList.remove( 'disabled' );
      return true;
    }
    else if ( flagCounter < numCases && !submitPredictionsButton.classList.contains( 'disabled' ) ) {
      submitPredictionsButton.classList.add( 'disabled' );
    }
    return false;
  }

  /**
   * Make sure that all table cases have been completed
   * @returns {boolean}
   */
  function validateTable() {
    for ( var i = 1; i <= window.simConfiguration.customization.applyCases.length; i++ ) {
      if ( !window.classroom.apply.casesComplete.has( i ) ) {
        return false;
      }
    }
    return true;
  }

  /**
   * Draw the specified case below the rules section of the apply tab
   * @param caseNumber
   */
  function drawCase( caseNumber ) {
    var outerCaseDiv = document.createElement( 'div' );
    outerCaseDiv.id = 'apply-case-' + caseNumber;
    outerCaseDiv.classList.add( 'apply-case' );

    var image = document.createElement( 'img' );
    image.src = 'img/apply-case' + caseNumber + '.png';
    image.classList.add( 'thumbnail' );
    outerCaseDiv.appendChild( image );

    var radioDiv = document.createElement( 'div' );
    window.createHypothesisRadioGroup( radioDiv, window.simConfiguration.hypothesisChoices, 'apply-case' + caseNumber );
    // Implement listener
    outerCaseDiv.appendChild( radioDiv );

    var textarea = document.createElement( 'textarea' );
    textarea.setAttribute( 'placeholder', 'Write your explanation here.' );
    textarea.id = 'apply-case-textarea' + caseNumber;
    // Implement listener
    outerCaseDiv.appendChild( textarea );
    $( outerCaseDiv ).hide();
    document.getElementById( 'apply-case-list' ).appendChild( outerCaseDiv );


    // Case specific radio button listener
    $( 'input[name=apply-case' + caseNumber + ']' ).click( function() {
      window.recordRadioButtonFired( this );
      window.classroom.apply.cases[ caseNumber - 1 ].prediction = this.value;
      validateFillOutCases();
      return true;
    } );

    // Case specific textarea listener
    $( textarea ).on( 'keyup change paste', function() {
      window.classroom.apply.cases[ caseNumber - 1 ].explanation = window.recordTextAreaChanged( this, window.classroom.apply.cases[ caseNumber - 1 ].explanation );
      validateFillOutCases();
      return true;
    } );
  }

  /**
   * Life cycle functions for this tab.
   * @type {{initialize: tabs.experiment.initialize, onLoad: tabs.experiment.onLoad, onLeave: tabs.experiment.onLeave, afterStateLoad: tabs.experiment.afterStateLoad}}
   */
  window.tabs.apply = {
    /**
     * Initial set up of the tab. Sets the click listeners and prepares the sim accordingly
     * Called from classroom-activity.switchTab.
     */
    initialize: function() {

      // Load in sim based text
      $( '#apply-text-1' ).text( window.simConfiguration.applyText1 );
      $( '#apply-text-2' ).text( window.simConfiguration.applyText2 );
      $( '#apply-text-3' ).text( window.simConfiguration.applyText3 );


      // Load reflect rules into apply rules box
      var $ari = $( '#apply-rules-input' );
      $ari.val( window.classroom.apply.rules );
      $ari.attr( 'disabled', 'disabled' );


      // Record rules box changes
      $ari.on( 'keyup change paste', function() {
        window.classroom.apply.rules = window.recordTextAreaChanged( this, window.classroom.apply.rules );
      } );


      // Create the cases at the bottom of the tab
      for ( var i = 1; i <= window.simConfiguration.customization.applyCases.length; i++ ) {
        // The closure is to set the click listener because it is lazily executed.
        (function() {
          var caseNumber = i;
          window.classroom.apply.cases.push( {
            prediction: '',
            explanation: '',
            winningFlagImgSrc: ''
          } );
          drawCase( caseNumber );
          var currentTab = $( '#apply-case-' + caseNumber + '-tab' );

          // Click listener for each case block
          currentTab.click( function() {
            $( '#apply-case-' + window.classroom.apply.currentCase ).hide();
            $( '#apply-case-' + caseNumber ).show();

            var $prev = $( '#apply-case-' + window.classroom.apply.currentCase + '-tab' );
            if ( $prev.hasClass( 'active' ) ) {
              $prev.removeClass( 'active' );
            }
            if ( !$( this ).hasClass( 'active' ) ) {
              $( this ).addClass( 'active' );
            }
            window.classroom.apply.currentCase = caseNumber;

            restoreCase( caseNumber, function() {
              window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setInactiveAndNotInteractive );
            } );
            window.recordButtonPressed( this );
          } );

        })( i );
      }

      window.classroom.apply.dataTable = new window.DataTable( recordBlankDataPoint, window.simIFrameClient,
        'classroomActivity', window.simConfiguration.applyTableHeaders, 'apply-table' );


      $( '#apply-predictions-button' ).click( function() {
        if ( validateFillOutCases() ) {

          $( '#apply-text-1' ).text( window.simConfiguration.applyText4 );
          $( '#apply-text-2' ).text( window.simConfiguration.applyText5 );
          $( '#apply-text-3' ).text( window.simConfiguration.applyText6 );

          window.classroom.apply.startedToTestCases = true;
          var tableElement = document.getElementById( 'data-table' );
          var dt = window.classroom.apply.dataTable;
          dt.setCellSetters( window.simConfiguration.getApplyCellSetters() );
          dt.drawTable( tableElement );

          // Create the five table rows.  These are only dependent on case number, not on dataPoints collected from the sim.
          window.simConfiguration.customization.applyCases.forEach( function() { window.classroom.apply.dataTable.record(); } );

          $( '#apply-case-list' ).hide();
          $( '#apply-submit-rules' ).show();
          $( this ).hide();

          window.simIFrameClient.invokeSequence(
            window.simConfiguration.customization.applyPredictions
              .concat( window.simConfiguration.customization.disablePullers )
              .concat( window.simConfiguration.customization.applyCases[ 0 ] )
          );
          window.classroom.apply.currentCase = 1;

          // Add class to highlight the first row.
          $( '#trialTableRow1' ).addClass( 'activeRow' );

          window.recordButtonPressed( this );
        }
        else {
          window.toast( 'Make sure to fill out all of the case predictions.' );
        }
      } );

      $( '#apply-edit-rules-button' ).click( function() {
        var $applyRulesInput = $( '#apply-rules-input' );
        var $applyEditRulesButton = $( '#apply-edit-rules-button' );
        if ( $applyRulesInput.attr( 'disabled' ) ) {
          $applyRulesInput.attr( 'disabled', null );
          $applyEditRulesButton.text( 'Save Rules' );
        }
        else {
          $applyEditRulesButton.text( 'Edit Rules' );
          $applyRulesInput.attr( 'disabled', 'disabled' );
        }
        window.recordButtonPressed( this );
      } );

      $( '#apply-submit-rules' ).click( function() {
        if ( validateTable() ) {
          if ( !window.classroom.apply.hasSubmittedBefore ) {
            window.classroom.apply.hasSubmittedBefore = true;
            window.toast( 'This is your final submission. Make sure your rules are complete and then click on Submit Rules again.' );

            window.classroom.apply.printingReport = true;


          }
          else {
            window.toast( 'Your final answer has been submitted.');

            // Make sim inactve and not interactive
            window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setInactiveAndNotInteractive );

            // Hide Data Table
            var tableElement = document.getElementById( 'data-table' );
            tableElement.innerHTML = '';

            // generateReport will open a popup and load the html there as well
            // as returning the html as a string in the callback.
            // This is for recording purposes.
            window.generateReport( window.drawReport(), window.simConfiguration.pageTitle + ' | Report',
              function( generatedHTML ) {
                window.recordHTMLReport( generatedHTML );
              } );
          }

          window.recordButtonPressed( this );
        }
        else {
          window.toast( 'Please test all the cases' );
        }
      } );


      // Restore the first case to begin
      restoreCase( 1, function() {
        window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setInactiveAndNotInteractive );
      } );
      window.classroom.apply.currentCase = 1;
      $( '#apply-case-1' ).show();
      $( '#apply-case-1-tab' ).addClass( 'active' );
      $( '#apply-submit-rules' ).hide();
      this.onLoad();

      var finalInterval = setInterval( function() {
        if ( validateTable() ) {
          clearInterval( finalInterval );
          // Allow the final submit button to be clicked
          $( '#apply-submit-rules' ).removeClass( 'disabled' );
          window.toast( 'You\'ve finished testing all cases! Review your rules and make any final changes then click "Submit Rules"' );

        }
      }, 500 );
    },

    /**
     * Sets the properties of the classroom object and customizations of the sim
     * that must be done each time traveling to this tab.
     * Called from classroom-activity.switchTab whenever going to this tab
     */
    onLoad: function() {
      window.simIFrameClient.invokeSequence( window.simConfiguration.getApplyOnLoadCustomizations()
        .concat( window.simConfiguration.customization.disablePullers )
      );

      if ( window.classroom.apply.startedToTestCases ) {
        var tableElement = document.getElementById( 'data-table' );
        tableElement.innerHTML = '';
        var dt = window.classroom.apply.dataTable;
        dt.setCellSetters( window.simConfiguration.getApplyCellSetters( dt ) );
        dt.drawTable( tableElement );
        $( tableElement ).show();

        // Click on the previously active row to highlight it
        $( '#' + window.classroom.apply.activeRowID ).addClass( 'activeRow' );
      }

    },
    /**
     * Called from classroom-activity.switchTab whenever leaving this tab and before loading/initializing the next
     */
    onLeave: function() {

      if ( window.classroom.apply.startedToTestCases ) {
        window.classroom.apply.activeRowID = $( '.activeRow' )[ 0 ].id;
      }

      // Clear the data table
      var tableElement = document.getElementById( 'data-table' );
      tableElement.innerHTML = '';
    },

    /**
     * Called from classroom-activity.switchTab after onLoad and this tab's customizations have been invoked and loaded
     */
    afterStateLoad: function() {}
  };
})();