// Copyright 2016, University of Colorado Boulder

(function() {
  'use strict';

  /**
   * Creates a list entry on the initial experiment screen for an idea.
   * All values default to blank.
   * If isOriginal is false, it does not create buttons and disables the textarea.
   *
   * @param ideaNumber
   * @param isOriginal {boolean}
   * @returns {Element} - div that holds the created idea list element
   */
  function createIdeaListElement( ideaNumber, isOriginal ) {
    var textarea = document.createElement( 'textarea' );

    if ( isOriginal ) {
      var buttonSpan = document.createElement( 'span' );
      var submitIdeaButton = document.createElement( 'button' );
      submitIdeaButton.setAttribute( 'type', 'button' ); // Prevent default page reload behavior
      submitIdeaButton.id = 'idea-submit-' + ideaNumber;
      submitIdeaButton.innerHTML = 'Submit Idea';
      submitIdeaButton.classList.add( 'phet-button' );
      submitIdeaButton.classList.add( 'disabled' );
      buttonSpan.appendChild( submitIdeaButton );

      var testIdeaButton = document.createElement( 'button' );
      testIdeaButton.setAttribute( 'type', 'button' ); // Prevent default page reload behavior
      testIdeaButton.id = 'idea-test-' + ideaNumber;
      testIdeaButton.innerHTML = 'Test Idea';
      testIdeaButton.setAttribute( 'style', 'display: none;' );
      testIdeaButton.classList.add( 'phet-button' );
      buttonSpan.appendChild( testIdeaButton );

      textarea.id = 'idea-text-' + ideaNumber;
      textarea.setAttribute( 'placeholder', 'Type your idea here.' );

      $( textarea ).on( 'keyup change paste', function() {

        if ( this.value ) {
          window.classroom.experiment.ideas[ ideaNumber - 1 ].idea =
            window.recordTextAreaChanged( this, window.classroom.experiment.ideas[ ideaNumber - 1 ].idea );
          if ( submitIdeaButton.classList.contains( 'disabled' ) ) {
            submitIdeaButton.classList.remove( 'disabled' );
          }
        }
        else {
          if ( !submitIdeaButton.classList.contains( 'disabled' ) ) {
            submitIdeaButton.classList.add( 'disabled' );
          }
        }
      } );

      $( submitIdeaButton ).click( function() {
        if ( !submitIdeaButton.classList.contains( 'disabled' ) ) {
          textarea.setAttribute( 'disabled', 'disabled' );
          $( '#idea-text-copy-' + ideaNumber ).val( textarea.value );
          $( submitIdeaButton ).hide();
          $( testIdeaButton ).show();
          $( '#idea-list-item-' + ideaNumber ).addClass( 'submitted' );

          window.recordButtonPressed( this );
        }
        else {
          window.toast( 'Please record an idea to test first.' );
        }
      } );

      $( testIdeaButton ).click( function() {
        // Switch the workspace
        $( '#idea-list' ).hide();
        $( '#idea-space-' + ideaNumber ).show();
        $( '#experiment-header' ).text( 'Collect Data' );

        window.classroom.experiment.currentIdeaIndex = ideaNumber - 1;
        window.classroom.experiment.inTest = true;

        window.drawDataTable( window.classroom.experiment.ideas[ window.classroom.experiment.currentIdeaIndex ].dataTable, 'experiment' );

        if ( !window.classroom.experiment.ideas[ window.classroom.experiment.currentIdeaIndex ].tested ) {
          // Activate the sim
          var activateSimSeq = window.simConfiguration.getIdeaCustomizations( ideaNumber );
          if ( window.classroom.experiment.ideas[ ideaNumber - 1 ].tested ) {
            activateSimSeq.push( { phetioID: 'phetio', method: 'setInteractive', args: [ false ] } );
          }
          // Restore state and set customization
          window.simIFrameClient.invokeSequence( activateSimSeq );

          // Clear greyscale
          $( '#simulation' ).css( 'filter', '' );
          $( '#done-testing-' + ideaNumber ).show();
        }
        else {
          // Always hide the save/edit buttons, this is especially for reviewing the test.
          $( '#save-summary-' + ideaNumber ).hide();
          $( '#edit-summary-' + ideaNumber ).hide();

          // When experiment isn't the active tab and you have finished testing this idea
          if ( window.classroom.activeTabId !== 'experiment' ) {
            $( '#phet-data-table-record').addClass('disabled');
          }
        }



        // Record in the data stream
        window.recordButtonPressed( this );
      } );
    }
    else {
      textarea.id = 'idea-text-copy-' + ideaNumber;
      textarea.setAttribute( 'disabled', 'disabled' );
    }

    var titleSpan = document.createElement( 'span' );
    titleSpan.innerHTML = 'Idea ' + ideaNumber;
    titleSpan.classList.add( 'idea-label' );

    var outerDiv = document.createElement( 'div' );
    outerDiv.appendChild( titleSpan );
    outerDiv.appendChild( textarea );
    if ( isOriginal ) {
      outerDiv.appendChild( buttonSpan );
      outerDiv.id = 'idea-list-item-' + ideaNumber;
    }
    else {
      outerDiv.classList.add( 'submitted' );
    }

    return outerDiv;
  }


  /**
   * Creates an HTML element to plan, test, and summarize an idea.
   * Creates listeners for all the button and textarea elements in this div.
   *
   * @param ideaNumber
   * @returns {Element} - div element that holds the ideaSpace
   */
  function createIdeaSpace( ideaNumber ) {
    var bulletString = '<span class="required">&bull;</span>';


    var directiveDiv = document.createElement( 'div' );
    directiveDiv.innerHTML = '<p>Test your ideas using the simulation.</p>';
    directiveDiv.id = 'idea-space-directive-' + ideaNumber;
    directiveDiv.classList.add( 'idea-space-directive' );

    var useRecordButtonBullet = document.createElement( 'p' );
    useRecordButtonBullet.innerHTML = bulletString + window.simConfiguration.experimentText2;
    directiveDiv.appendChild( useRecordButtonBullet );
    var organizeDataBullet = document.createElement( 'p' );
    organizeDataBullet.innerHTML = bulletString + window.simConfiguration.experimentText3;
    directiveDiv.appendChild( organizeDataBullet );


    var ideaDiv = createIdeaListElement( ideaNumber, false );

    // This span will have the record data button added to it in the future.
    var recordingDiv = document.createElement( 'div' );
    recordingDiv.id = 'recording-div-' + ideaNumber;
    recordingDiv.classList.add( 'recording-div' );


    var whatILearnedSpan = document.createElement( 'span' );
    whatILearnedSpan.innerHTML = 'Use evidence from your data table in your summary.';

    var whatILearnedText = document.createElement( 'textarea' );
    whatILearnedText.id = 'idea-summary-' + ideaNumber;
    whatILearnedText.setAttribute( 'placeholder', 'Summarize what you learned here.' );

    var saveWhatILearnedButton = document.createElement( 'button' );
    saveWhatILearnedButton.id = 'save-summary-' + ideaNumber;
    saveWhatILearnedButton.setAttribute( 'type', 'button' );
    saveWhatILearnedButton.innerHTML = 'Save Summary';
    saveWhatILearnedButton.classList.add( 'phet-button' );
    saveWhatILearnedButton.classList.add( 'disabled' );

    var editWhatILearnedButton = document.createElement( 'button' );
    editWhatILearnedButton.id = 'edit-summary-' + ideaNumber;
    editWhatILearnedButton.setAttribute( 'type', 'button' );
    editWhatILearnedButton.innerHTML = 'Edit Summary';
    editWhatILearnedButton.classList.add( 'phet-button' );
    $( editWhatILearnedButton ).hide();

    var summaryButtonSpan = document.createElement( 'span' );
    summaryButtonSpan.appendChild( saveWhatILearnedButton );
    summaryButtonSpan.appendChild( editWhatILearnedButton );

    var whatILearnedDiv = document.createElement( 'div' );
    whatILearnedDiv.appendChild( whatILearnedSpan );
    whatILearnedDiv.appendChild( whatILearnedText );
    whatILearnedDiv.appendChild( summaryButtonSpan );


    var testAnotherIdeaButton = document.createElement( 'button' );
    testAnotherIdeaButton.setAttribute( 'type', 'button' );
    testAnotherIdeaButton.innerHTML = 'Test Another Idea';
    testAnotherIdeaButton.classList.add( 'phet-button' );
    testAnotherIdeaButton.classList.add( 'finish-test' );

    var finishDiv = document.createElement( 'div' );
    finishDiv.appendChild( testAnotherIdeaButton );
    finishDiv.id = 'idea-finish-' + ideaNumber;

    var spaceDiv = document.createElement( 'div' );
    spaceDiv.id = 'idea-space-' + ideaNumber;
    spaceDiv.classList.add( 'idea-space' );
    spaceDiv.setAttribute( 'style', 'display: none;' );
    spaceDiv.appendChild( directiveDiv );
    spaceDiv.appendChild( ideaDiv );
    spaceDiv.appendChild( recordingDiv );
    spaceDiv.appendChild( whatILearnedDiv );
    spaceDiv.appendChild( finishDiv );

    document.getElementById( 'experiment-workspace' ).appendChild( spaceDiv );

    $( whatILearnedText ).on( 'keyup change paste', function() {
      window.classroom.experiment.ideas[ ideaNumber - 1 ].summary =
        window.recordTextAreaChanged( this, window.classroom.experiment.ideas[ ideaNumber - 1 ].summary );
      if ( this.value ) {
        if ( saveWhatILearnedButton.classList.contains( 'disabled' ) ) {
          saveWhatILearnedButton.classList.remove( 'disabled' );
        }
      }
      else {
        if ( !saveWhatILearnedButton.classList.contains( 'disabled' ) ) {
          saveWhatILearnedButton.classList.add( 'disabled' );
        }
      }

    } );

    $( saveWhatILearnedButton ).click( function() {
      if ( !saveWhatILearnedButton.classList.contains( 'disabled' ) ) {

        // Disable text
        whatILearnedText.setAttribute( 'disabled', 'disabled' );

        // Hide this button
        $( saveWhatILearnedButton ).hide();
        $( editWhatILearnedButton ).show();

        window.classroom.experiment.ideas[ ideaNumber - 1 ].whatILearnedSubmitted = true;


        window.recordButtonPressed( this );

      }
      else {
        window.toast( 'Please summarize what you learned first.' );
      }
    } );

    $( editWhatILearnedButton ).click( function() {
      $( editWhatILearnedButton ).hide();
      $( saveWhatILearnedButton ).show();
      $( whatILearnedText ).prop( 'disabled', false );
      window.classroom.experiment.ideas[ ideaNumber - 1 ].whatILearnedSubmitted = false;

    } );

    $( testAnotherIdeaButton ).click( function() {
      if ( !testAnotherIdeaButton.classList.contains( 'disabled' ) ) {
        // get the state and save it
        window.simIFrameClient.invoke( 'phetio', 'getState', [], function( oldState ) {
          var currentIdea = window.classroom.experiment.ideas[ ideaNumber - 1 ];
          currentIdea.simState = oldState;

          // Reset the sim state
          window.simIFrameClient.invokeSequence( [
            { phetioID: window.simConfiguration.modelTandem, method: 'reset' },
            { phetioID: window.simConfiguration.activePropertyTandem, method: 'setValue', args: [ false ] }
          ] );
          // Reset the greyscale filter
          $( '#simulation' ).css( 'filter', 'grayscale(100%)' );


          // Clear the data table
          document.getElementById( 'data-table' ).innerHTML = '';

          // Save whatILearned text
          // $( '#save-summary-' + ideaNumber ).click();

          // Delete the record button
          recordingDiv.removeChild( document.getElementById( 'phet-data-table-record' ) );

          if ( currentIdea.whatILearnedSubmitted ) {
            currentIdea.tested = true;
            // Set every time, but the first one is all that matters.
            window.classroom.experiment.firstTestCompleted = true;
          }
          // Change the button to view this idea
          if ( currentIdea.tested ) {
            $( '#idea-test-' + ideaNumber ).text( 'Review Test' );
          }
          else {
            // Change the button to view this idea
            $( '#idea-test-' + ideaNumber ).text( 'Continue Test' );
          }

          window.classroom.experiment.inTest = false;

          // Switch back to idea-list space
          $( '#idea-space-' + ideaNumber ).hide();
          $( '#idea-list' ).show();
          $( '#experiment-header' ).text( 'Experiment' );


          // If the first test has been completed and the next button isn't already displayed
          // (meaning that we already showed the no-new-ideas button and it was clicked on)
          if ( window.classroom.experiment.firstTestCompleted && $( '#experiment-next' ).css( 'display' ) === 'none' ) {
            $( '#no-new-ideas' ).show();
          }

          window.recordButtonPressed( this );
        } );
      }
      else {
        window.toast( 'Please complete the plan and summary.' );
      }
    } );


    return spaceDiv;
  }

  /**
   * Checks that less than 5 ideas already exist.
   * If not, creates a new entry in the list of ideas, and a new div to show the planning and summary stages for that idea.
   * Adds a new idea object to window.classroom.experiment.ideas
   */
  function createIdea() {
    var ideaNumber = window.classroom.experiment.ideas.length + 1;
    if ( ideaNumber === 5 ) {
      $( '#add-idea' ).hide();
    }
    if ( ideaNumber > 5 ) {
      return false;
    }
    window.classroom.experiment.ideas.push( {
      idea: '',
      plan: '',
      summary: '',
      // simState: {}, // Get sim state?
      tested: false,
      whatILearnedSubmitted: false,
      dataTable: new window.DataTable( window.simConfiguration.recordDataPoint, window.simIFrameClient,
        'classroomActivity', window.simConfiguration.experimentTableHeaders, 'experiment-table' )
    } );

    var experimentNext = document.getElementById( 'experiment-next' );
    document.getElementById( 'ideas-to-test' ).appendChild( createIdeaListElement( ideaNumber, true ) );
    document.getElementById( 'experiment-workspace' ).insertBefore( createIdeaSpace( ideaNumber ), experimentNext );
  }

  /**
   * Life cycle functions for this tab.
   * @type {{initialize: tabs.experiment.initialize, onLoad: tabs.experiment.onLoad, onLeave: tabs.experiment.onLeave, afterStateLoad: tabs.experiment.afterStateLoad}}
   */
  window.tabs.experiment = {
    /**
     * Initial set up of the tab. Sets the click listeners and prepares the sim accordingly
     * Called from classroom-activity.switchTab.
     */
    initialize: function() {

      createIdea();
      createIdea();


      $( '#add-idea' ).click( function() {
        createIdea();
        window.recordButtonPressed( this );
      } );

      $( '#no-new-ideas' ).click( function() {
        $( '#no-new-ideas' ).hide();
        $( '#experiment-next' ).show();
        $( '#reflect' ).toggleClass( 'disabled pickable' );
        $( '#add-idea' ).hide();
        window.recordButtonPressed( this );
      } );

      $( '#experiment-text' ).text( window.simConfiguration.experimentText1 );
      window.classroom.experiment.inTest = false;

      this.onLoad();
    },

    /**
     * Sets the properties of the classroom object and customizations of the sim
     * that must be done each time traveling to this tab.
     * Called from classroom-activity.switchTab whenever going to this tab
     */
    onLoad: function() {

      // If you are in a test at all draw the table
      if ( window.classroom.experiment.inTest ) {
        $( '#simulation' ).css( 'filter', '' );
        window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setActiveAndInteractive );
        var currentIdeaIndex = window.classroom.experiment.currentIdeaIndex;
        window.drawDataTable( window.classroom.experiment.ideas[ currentIdeaIndex ].dataTable, 'experiment' );
        if ( window.classroom.activeTabId !== 'experiment' && window.classroom.experiment.ideas[ currentIdeaIndex ].tested) {
          $( '#phet-data-table-record').addClass('disabled');
          $( '#simulation' ).css( 'filter', 'grayscale(100%)' );
        }
      }
      else {
        // Make the sim inactive and greyed out
        $( '#simulation' ).css( 'filter', 'grayscale(100%)' );
        window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setInactiveAndNotInteractive );

        // Draw a blank data table to give example of what will be tested.
        if (window.classroom.activeTabId === 'experiment' && window.classroom.experiment.firstTestCompleted){
          window.drawDataTable( window.classroom.experiment.ideas[ 0 ].dataTable, 'blankExperiment' );
        }
      }
    },

    /**
     * Called from classroom-activity.switchTab whenever leaving this tab and before loading/initializing the next
     */
    onLeave: function() {
      // Delete record button if you are in a test, it will be redrawn each time drawDataTable is called
      if ( window.classroom.experiment.inTest ) {
        var currentIdeaNumber = window.classroom.experiment.currentIdeaIndex + 1;
        var recordDiv = document.getElementById( 'recording-div-' + ( currentIdeaNumber) );
        recordDiv.removeChild( document.getElementById( 'phet-data-table-record' ) );

        if ( window.classroom.experiment.ideas[ currentIdeaNumber - 1 ].tested ) {
          // Save the 'whatILearned' text
          $( '#save-summary-' + currentIdeaNumber ).click();
          $( '#edit-summary-' + currentIdeaNumber ).hide();

        }
      }
      $( '#simulation' ).css( 'filter', '' );

      $( '#data-table' ).html('');
    },

    /**
     * Called from classroom-activity.switchTab after onLoad and this tab's customizations have been invoked and loaded
     */
    afterStateLoad: function() {
      var currentIdea = window.classroom.experiment.ideas[ window.classroom.experiment.currentIdeaIndex ];
      if ( window.classroom.experiment.inTest && !currentIdea.tested ) {
        //  should be active
        window.simIFrameClient.invokeSequence( window.simConfiguration.customization.setActiveAndInteractive
          .concat( window.simConfiguration.customization.enablePullers ) );
      }
    }
  };
})();