// Copyright 2016, University of Colorado Boulder

/**
 * Provides support for Recording data from a PhET-iO Simulation.  Data can be recorded to:
 * - Metacog
 * - console
 * - localFile (saves a log file in downloads)
 *
 * It is configured through query parameters.
 *
 * It has the following dependencies:
 *  metacog.js
 *  jquery.js (required only by metacog)
 *  assert.js
 *  ../../js/SimIFrameClient.js
 *  ../../js/WrapperUtils.js
 *
 *  @author Sam Reid (PhET Interactive Simulations)
 */
(function() {
  'use strict';

  // constants
  // Identify local data collection tests for when metadata (like session_id) is not retrieved from metacog.
  // This string is also hardcoded in Session.js
  var LOCAL_TEST_VALUE = 'localTest';

  /**
   * @constructor
   */
  window.Recorder = function() {

    // Flags for where to record to
    var recordTo = QueryStringMachine.getAll( {
      'console': { type: 'flag' },
      'metacog': { type: 'flag' },
      'localFile': { type: 'flag' }
    } );

    // Configure for metacog, if that query parameter is supplied
    var metacogQueryParameters = null;
    if ( recordTo.metacog ) {

      // All required
      metacogQueryParameters = QueryStringMachine.getAll( {

        // Basically the username for the Metacog application
        publisher_id: {
          type: 'string'
        },

        // Key word that can be used to categorize/search on this recorded session.
        widget_id: {
          type: 'string'
        },

        // The page the user used to submit the learner_id
        loginURL: {
          type: 'string'
        },

        // The name of the key that the learner_token service will use to authenticate with metacog
        key_name: {
          type: 'string'
        },

        // Show a debug console for the metacog api. This will show the events logged by the Metacog client.
        debugMetacog: {
          type: 'flag'
        }
      } );

      var learner_id = sessionStorage.getItem( 'phet.metacog.learner_id' );
      sessionStorage.removeItem( 'phet.metacog.learner_id' );

      if ( !learner_id ) {

        // The start function is still called, we must make sure it does nothing.
        this.start = function() {};

        // Send the user to the login page
        window.location.href = decodeURIComponent( metacogQueryParameters.loginURL );

        // No further configuration of the Recorder.
        return;
      }
    }

    // Display on the console where the messages are going, for debugging
    var outputs = [];
    recordTo.metacog && outputs.push( 'metacog' );
    recordTo.console && outputs.push( 'console' );
    recordTo.localFile && outputs.push( 'localFile' );

    console.log( 'Recorder initialized, sending data to: ', outputs );

    // Initialize localFile recording
    recordTo.localFile && saveToFileSetup();

    this.recordMessage = function( message ) {
      var parsedMessage = JSON.parse( message );
      if ( recordTo.console ) {
        console.log( JSON.stringify( parsedMessage, null, 2 ) );
      }
      if ( recordTo.localFile ) {
        window.eventArray.push( message );
      }

      // Send the data to Metacog
      recordTo.metacog && Metacog.send( {
        event: parsedMessage.phetioID + '.' + parsedMessage.event,
        data: parsedMessage,
        type: Metacog.EVENT_TYPE.MODEL
      } );
    };

    this.start = function( callback ) {

      var random8DigitString = Math.random().toFixed( 8 ).substring( 2 );
      var sessionID = random8DigitString + '@' + Date.now();

      // Configure Metacog for messaging
      if ( recordTo.metacog ) {

        var xhr = new XMLHttpRequest();
        xhr.onreadystatechange = function() {
          var DONE = 4; // readyState 4 means the request is done.
          var OK = 200; // status 200 is a successful return.
          if ( xhr.readyState === DONE ) {
            if ( xhr.status === OK ) {
              console.log( xhr.responseText );
              var learner_token = JSON.parse( xhr.responseText ).learner_token;
              console.log( 'recording to sessionID: ' + sessionID );
              var sessionMetadata = {
                'publisher_id': metacogQueryParameters.publisher_id,
                'widget_id': metacogQueryParameters.widget_id,
                'learner_token': learner_token,
                'learner_id': learner_id,
                'session_id': sessionID
              };
              var config = {
                session: sessionMetadata,
                log_tab: metacogQueryParameters.debugMetacog,
                mode: 'production'
              };

              Metacog.init( config, function() {

                window.addEventListener( 'unload', function() {
                  console.log( 'Stopping Metacog' );
                  Metacog.stop();
                } );
                Metacog.start();

                // Load the content after Metacog is ready to go, so that no messages are lost
                callback( sessionMetadata );
              } );
            } // 'This is the returned text.'

            else {
              console.log( 'Error: ' + xhr.status ); // An error occurred during the request.
            }
          }
        };

        var learnerTokenServiceURL = 'https://phet-io.colorado.edu/learner-token-service' +
                                     '?publisher_id=' + metacogQueryParameters.publisher_id +
                                     '&key_name=' + metacogQueryParameters.key_name +
                                     '&learner_id=' + learner_id;
        xhr.open( 'GET', learnerTokenServiceURL );
        xhr.send( null );
      }
      else {
        // Return session metadata that is not specific to Metacog because we aren't recording to them.
        callback( { 'session_id': sessionID } );
      }
    };
  };

  /**
   * Initialize recording to localFile
   * This assigns window.saveLogToFile, the local saving method in the browser
   */
  var saveToFileSetup = function() {
    // General instructions about how to use localFile log saving.
    console.log( 'Recording to localFile has begun.' );
    console.log( 'When you are finished recording, save the data log with this command in your browser console: \n' +
                 'window.saveLogToFile("filename");' );

    // Initialize the main event array that will be saved later
    window.eventArray = [];


    // Make the events look just like the ones that Metacog returns so that we can more accurately test the playback parser/wrapper.
    var formatLogFile = function( events ) {
      var metacogLikeIndex = -1;
      var metacogLikeEvents = events.map( function( event ) {
        var eventJSON = JSON.parse( event );
        metacogLikeIndex++;
        return {
          'event': 'testing.' + eventJSON.event,
          'timestamp': Date.now(),
          'index': metacogLikeIndex,
          'type': 'test',
          'data': eventJSON
        };


      } );
      var sessionString = {
        'session': {
          'session_id': LOCAL_TEST_VALUE,
          'publisher_id': LOCAL_TEST_VALUE,
          'application_id': LOCAL_TEST_VALUE,
          'widget_id': LOCAL_TEST_VALUE,
          'learner_id': LOCAL_TEST_VALUE
        }
      };
      sessionString.events = metacogLikeEvents;
      return JSON.stringify( sessionString ).replace( new RegExp( '[\n\r]', 'g' ), '' ); //eslint-disable-line no-control-regex
    };

    /**
     * Function to save data to a file
     * @param file {Blob}
     * @param filename {String}
     */
    var saveFile = function( file, filename ) {
      var a = document.createElement( 'a' );
      if ( window.navigator.msSaveOrOpenBlob ) { // IE10+
        window.navigator.msSaveOrOpenBlob( file, filename );
      }
      else { // Others
        var url = URL.createObjectURL( file );
        a.href = url;
        a.download = filename;
        document.body.appendChild( a );
        a.click();
        setTimeout( function() {
          document.body.removeChild( a );
          window.URL.revokeObjectURL( url );
        }, 0 );
      }
    };

    /**
     * Save the event array to a log file that will be downloaded by the browser.
     * NOTE: On Safari this will only open a new tab with the json data there to manually save.
     * @param filename - '.log' added to the end of it
     */
    window.saveLogToFile = function( filename ) {
      var logText = formatLogFile( window.eventArray );
      var blob = new Blob( [ logText ], { type: 'text/plain;charset=utf-8' } );
      saveFile( blob, filename + '.log' );
    };
  };
})();