// Copyright 2016, University of Colorado Boulder

/**
 * Play back a single recorded event into a live simulation.
 *
 * @param {Object} event - event json object representing the event
 * @param {SimIFrameClient} simIFrameClient - where the playback commands will be applied
 * @param {function} callback - once the event is complete, the callback is called
 */
window.handlePlaybackEvent = function( event, simIFrameClient, callback ) {
  'use strict';

  var invokeSetDisplaySize = function( eventObject ) {
    simIFrameClient.invoke(
      'phetio',
      'setDisplaySize',
      [ eventObject.parameters.width, eventObject.parameters.height ],
      callback );
  };

  if ( event.event === 'simStarted' ) {

    // capture the metadata parameters, so we will know important things like the browser userAgent, etc.
    var parameters = event.parameters;
    console.log( parameters );
  }

  // The startup resize event is a child of simStarted, we must detect it so that the initial screen size is correct
  if ( event.children ) {

    for ( var i = 0; i < event.children.length; i++ ) {
      var obj = event.children[ i ];
      if ( obj.event === 'displaySize' ) {
        invokeSetDisplaySize( obj );
        return;
      }
    }
  }

  if ( event.event === 'stepSimulation' ) {
    simIFrameClient.invoke( 'phetio', 'stepSimulation', [ event.parameters.dt ], callback );
  }
  else if ( event.event === 'inputEvent' ) {

    // Play it to the downstream sim.
    simIFrameClient.invoke( 'phetio', 'invokeInputEvent', [ event.parameters.string ], callback );
  }
  else if ( event.event === 'displaySize' ) {
    invokeSetDisplaySize( event );
  }
  else if ( event.event === 'invoked' ) {
    // Don't invoke those methods because the callbacks don't exist anymore
    if ( event.parameters.args && JSON.stringify( event.parameters.args ).indexOf( 'functionHandle@messageID=' ) >= 0 ) {
      callback();
      return;
    }

    if ( event.parameters.method === 'launchSimulation' ) {

      // Proceed with playback once the simulation has fully launched
      simIFrameClient.onSimInitialized( callback );
      callback = function() { };
    }
    if ( !event.parameters.phetioID ) {

      // Iterate over all children for a composite message.
      if ( event.parameters.messages ) {
        var messages = [];
        for ( var k = 0; k < event.parameters.messages.length; k++ ) {
          var message = event.parameters.messages[ k ];
          console.log( 'invoking children: ', message.phetioID + '.' + message.method, JSON.stringify( message.args ) );
          messages.push( { phetioID: message.phetioID, method: message.method, args: message.args } );
        }
        simIFrameClient.invokeSequence( messages, function() {
          callback();
        } );
      }
    }
    else {
      console.log( 'invoking ', event.parameters.phetioID + '.' + event.parameters.method, JSON.stringify( event.parameters.args ) );
      simIFrameClient.invoke( event.parameters.phetioID, event.parameters.method, event.parameters.args, callback );
    }
  }
  else {
    callback();
  }
};