// Copyright 2015-2017, University of Colorado Boulder

/**
 * This file is responsible for creating the custom ui component divs for each tandem id. These are specific to the type
 * and have built in functions to support the features of each type. For example, a tandem of type TText will have a div
 * that includes a textbox that will change the value of the TText in the sim. TNodes include a visibility checkbox.
 */
(function() {
  'use strict';

  /**
   * Get the method from a type. Also searches supertypes
   * @param {string} methodName
   * @param {Object} type - type api
   * @returns {*}
   */
  var getMethod = function( methodName, type ) {
    if ( type[ methodName ] ) {
      return type[ methodName ];
    }
    else {
      if ( type.supertype ) {
        return getMethod( methodName, type.supertype );
      }
      else {
        return null;
      }
    }
  };

  /**
   * @param {string} str
   * @returns {string}
   */
  var uppercaseFirst = function( str ) {
    if ( str.length === 0 ) {
      return str;
    }
    else if ( str.length === 1 ) {
      return str.toUpperCase();
    }
    else {
      return str.charAt( 0 ).toUpperCase() + str.substring( 1 );
    }
  };

  /**
   * @param string
   * @returns {string}
   */
  var toWhitespaced = function( string ) {
    var result = uppercaseFirst( string.replace( /([A-Z])/g, function( match ) {return ' ' + uppercaseFirst( match.toLowerCase() );} ) );
    if ( result.toLowerCase() === 'phetio' ) {
      return 'PhETiO';
    }
    return result;
  };

  /**
   * Get the display name from a lower case, dot separated string.
   * @param {string} phetioID
   * @returns {*}
   */
  var getFriendlyName = function( phetioID ) {
    var array = phetioID.split( '.' );
    var tail = array[ array.length - 1 ];
    return toWhitespaced( tail );
  };

  /**
   * Creates an HTML span Element
   * @param {string} label
   * @param {Object} [options]
   * @returns {HTMLElement}
   */
  var createSpan = function( label, options ) {
    var span = document.createElement( 'span' );
    span.innerHTML = _.escape( label );
    if ( options && options.className ) {
      span.className = options.className;
    }
    return span;
  };

  /**
   * Creates an HTML br Element
   * @param {object} [options]
   * @returns {HTMLElement}
   */
  var createBR = function( options ) {
    var element = document.createElement( 'br' );
    if ( options && options.className ) {
      element.className = options.className;
    }
    return element;
  };

  /**
   * Creates an HTML input Element
   * @param {object} [options]
   * @returns {HTMLElement}
   */
  var createInput = function( options ) {
    var input = document.createElement( 'input' );
    input.type = options.type || 'checkbox';
    if ( options.id ) {
      input.id = options.id;
    }
    if ( options.checked ) {
      input.checked = options.checked;
    }
    if ( options.className ) {
      input.className = options.className;
    }
    if ( options.disabled ) {
      input.disabled = true;
    }
    input.checked = true; // TODO: correctly sync visible
    return input;
  };

  /**
   * Append the necessary components to an HTMLElement and wire those components up to control the "pickable"
   * functionality of a component in the simulation.
   * @param {SimIFrameClient} simIFrameClient
   * @param {Object} instance
   * @param {String} phetioID
   * @param {HTMLElement} div
   * @param {Object} options
   */
  var createPickableControls = function( simIFrameClient, instance, phetioID, div, options ) {
    div.appendChild( createSpan( 'pickable' ) );
    var pickableFalseCheckBox = createInput( {
      type: 'radio',
      id: 'checkbox_pickable_false_' + phetioID,
      checked: true,
      disabled: options.phetioReadOnly
    } );
    div.appendChild( pickableFalseCheckBox );
    div.appendChild( createSpan( 'false' ) );

    var pickableNullCheckBox = createInput( {
      type: 'radio',
      id: 'checkbox_pickable_null_' + phetioID,
      checked: true,
      disabled: options.phetioReadOnly
    } );
    div.appendChild( pickableNullCheckBox );
    div.appendChild( createSpan( 'null' ) );

    var pickableTrueCheckBox = createInput( {
      type: 'radio',
      id: 'checkbox_pickable_true_' + phetioID,
      checked: true,
      disabled: options.phetioReadOnly
    } );
    div.appendChild( pickableTrueCheckBox );
    div.appendChild( createSpan( 'true' ) );


    pickableFalseCheckBox.addEventListener( 'click', function() {
      simIFrameClient.invoke( phetioID, 'setPickable', [ false ] );
    } );
    pickableNullCheckBox.addEventListener( 'click', function() {
      simIFrameClient.invoke( phetioID, 'setPickable', [ null ] );
    } );
    pickableTrueCheckBox.addEventListener( 'click', function() {
      simIFrameClient.invoke( phetioID, 'setPickable', [ true ] );
    } );

    simIFrameClient.invoke( phetioID, 'isPickable', [], function( pickable ) {
      pickableFalseCheckBox.checked = false;
      pickableNullCheckBox.checked = false;
      pickableTrueCheckBox.checked = false;
      if ( pickable === false ) {
        pickableFalseCheckBox.checked = true;
      }
      else if ( pickable === null ) {
        pickableNullCheckBox.checked = true;
      }
      else if ( pickable === true ) {
        pickableTrueCheckBox.checked = true;
      }
      else {
        throw new Error( 'unknown value for pickable' );
      }
    } );

    instance.pickableFalseCheckBox = pickableFalseCheckBox;
    instance.pickableNullCheckBox = pickableNullCheckBox;
    instance.pickableTrueCheckBox = pickableTrueCheckBox;


    // Listen for pickability changes in the node and update the checkboxes accordingly.
    simIFrameClient.invoke( phetioID, 'addPickableListener', [ function( pickable ) {
      instance.pickableFalseCheckBox.checked = pickable === false;
      instance.pickableNullCheckBox.checked = pickable === null;
      instance.pickableTrueCheckBox.checked = pickable === true;
    } ] );
  };

  /**
   * Creates a div specific to the TType. This involves lots of conditional logic to add specific components to enable
   * proper customization for the instance.
   * @param {string} phetioID - tandem id
   * @param {Object} type - the TType api
   * @param {Function} isType
   * @param {SimIFrameClient} simIFrameClient
   * @param {Object} stateObject
   * @param {Object} options - see phetio.addInstance
   * @returns {{typeName: string, type: {Object}, div: HTMLElement}}
   */
  window.createInstanceProxyDiv = function( phetioID, type, isType, simIFrameClient, stateObject, options ) {
    var typeName = type.typeName;
    var div = document.createElement( 'div' );

    div.appendChild( createSpan( getFriendlyName( phetioID ), { className: 'friendlyName' } ) );
    div.appendChild( createBR() );
    div.appendChild( createSpan( phetioID + ': ' + typeName, { className: 'codeName' } ) );
    div.appendChild( createBR( { className: 'codeNameBR' } ) );

    if ( options.phetioInstanceDocumentation && options.phetioInstanceDocumentation.length > 0 ) {
      div.appendChild( createSpan( options.phetioInstanceDocumentation, { className: 'instanceDoc' } ) );
      div.appendChild( createBR() );
    }
    if ( options.phetioReadOnly ) {
      div.appendChild( createSpan( 'Read Only: do not try to set this component' ) );
      div.appendChild( createBR() );
    }
    div.appendChild( createSpan( 'Initial State: ' + JSON.stringify( stateObject, null, 2 ) ) );
    div.appendChild( createBR() );

    var instance = { typeName: typeName, type: type, div: div };
    var input = null;

    if ( isType( 'TNode', type ) ) {

      var visibleCheckBox = createInput( {
        id: 'checkbox_' + phetioID,
        checked: true,
        disabled: options.phetioReadOnly
      } );

      div.appendChild( visibleCheckBox );

      div.appendChild( createSpan( 'visible' ) );

      var detachButton = document.createElement( 'button' );
      detachButton.innerHTML = 'Detach';

      // TODO: figure out detachment
      // div.appendChild( detachButton );

      detachButton.addEventListener( 'click', function() {
        simIFrameClient.invoke( phetioID, 'detach', [], function() {
          detachButton.pressed = true;
        } );
      } );
      div.appendChild( createBR() );

      createPickableControls( simIFrameClient, instance, phetioID, div, options );

      div.appendChild( createBR() );

      visibleCheckBox.addEventListener( 'click', function() {
        simIFrameClient.invoke( phetioID, 'setVisible', [ visibleCheckBox.checked ] );
      } );

      // Initial Values
      simIFrameClient.invoke( phetioID, 'isVisible', [], function( visible ) {
        visibleCheckBox.checked = visible;
      } );

      instance.customizable = true;
      instance.visibleCheckBox = visibleCheckBox;
      instance.detachButton = detachButton;

      simIFrameClient.invoke( phetioID, 'addVisibleListener', [ function( visible ) {
        visibleCheckBox.checked = visible;
      } ] );
    }

    if ( isType( 'TImage', type ) ) {

      var fileButton = createInput( { type: 'file', disabled: options.phetioReadOnly } );
      div.appendChild( fileButton );

      $( fileButton ).change( function( e ) {

        for ( var i = 0; i < e.originalEvent.target.files.length; i++ ) {

          var file = e.originalEvent.target.files[ i ];

          var reader = new FileReader();
          reader.onloadend = function() {
            // img.src = reader.result;
            simIFrameClient.invoke( phetioID, 'setImage', [
              reader.result
            ] );
            instance.base64 = reader.result;
          };
          reader.readAsDataURL( file );
        }
      } );
    }

    if ( isType( 'TPhetButton', type ) ) {
      div.appendChild( createBR() );
    }
    if ( isType( 'TProperty', type ) && !isType( 'TDerivedProperty', type ) && type.typeAPI.valueType.typeName === 'TBoolean' ) {
      input = createInput( { id: 'checkbox_' + phetioID, disabled: options.phetioReadOnly } );
      div.appendChild( input );
      div.appendChild( createSpan( 'selected' ) );
      div.appendChild( createBR() );

      input.addEventListener( 'click', function() {
        simIFrameClient.invoke( phetioID, 'setValue', [ input.checked ] );
      } );

      simIFrameClient.invoke( phetioID, 'link', [ function( checked ) {
        input.checked = checked;
      } ] );

      instance.customizable = true;
    }
    if ( isType( 'TProperty', type ) && !isType( 'TDerivedProperty', type ) && type.typeAPI.valueType.typeName === 'TNumber' ) {
      // TODO: Distinguish between integer/floating point/ranges/etc.
      input = document.createElement( 'input' );
      input.type = 'range';
      input.name = 'points';
      var min = 0;
      var max = 1;

      if ( stateObject.range ) {
        min = stateObject.range.min;
        var unitsString = stateObject.units || '';
        div.appendChild( createSpan( 'min: ' + min + ' ' + unitsString ) );
        div.appendChild( createBR() );

        max = stateObject.range.max;
        div.appendChild( createSpan( 'max: ' + max + ' ' + unitsString ) );
        div.appendChild( createBR() );
      }
      if ( stateObject.validValues ) {
        div.appendChild( createSpan( 'possible values: ' + stateObject.validValues ) );
        div.appendChild( createBR() );
      }

      // Edge cases for infinity
      if ( typeof min === 'string' && min.indexOf( 'INFINITY' ) >= 0 ) {
        min = -100;
      }
      if ( typeof max === 'string' && max.indexOf( 'INFINITY' ) >= 0 ) {
        max = 100;
      }
      input.min = min + '';
      input.max = max + '';

      // integral sliders should only try to set integer values
      if ( stateObject.valueType === 'Integer' ) {
        input.step = 1;
      }
      else {
        input.step = (max - min) / 100 + '';
      }

      var onlyHasOneValue = false;
      if ( max === min ) {
        onlyHasOneValue = true;
      }
      if ( stateObject.validValues && stateObject.validValues.length === 1 ) {
        onlyHasOneValue = true;
      }
      if ( !onlyHasOneValue && !stateObject.validValues && !options.phetioReadOnly ) {
        div.appendChild( input );
        div.appendChild( createBR() );
      }
      else if ( !onlyHasOneValue && stateObject.validValues ) {

        // create one button for each value
        stateObject.validValues.forEach( function( value ) {

          var button = document.createElement( 'button' );
          button.innerHTML = 'Set to ' + value;
          !options.phetioReadOnly && div.appendChild( button );

          button.addEventListener( 'click', function() {
            simIFrameClient.invoke( phetioID, 'setValue', [ value ], function( text ) {
              input.value = text;
            } );
          } );
        } );
        !options.phetioReadOnly && div.appendChild( createBR() );
      }
      var update = function() {
        simIFrameClient.invoke( phetioID, 'setValue', [ parseFloat( input.value ) ] );
      };
      input.oninput = update;
      input.onchange = update;

      simIFrameClient.invoke( phetioID, 'link', [ function( value ) {
        input.value = value;
      } ] );

      instance.customizable = true;

    }
    if ( isType( 'TProperty', type ) ) {
      var valueSpan = createSpan( 'value', { className: 'propertyValue' } );
      div.appendChild( valueSpan );
      div.appendChild( createBR() );

      simIFrameClient.invoke( phetioID, 'link', [ function( newValue ) {
          valueSpan.innerHTML = 'Current Value: ' + JSON.stringify( newValue );
        } ]
      );
    }
    if ( typeName === 'TCheckBox' ) {
      input = createInput( { id: 'checkbox_' + phetioID, disabled: options.phetioReadOnly } );
      !options.phetioReadOnly && div.appendChild( input );

      div.appendChild( createSpan( 'selected' ) );

      input.addEventListener( 'click', function() {
        simIFrameClient.invoke( phetioID, 'setChecked', [ input.checked ] );
      } );

      simIFrameClient.invoke( phetioID, 'link', [ function( checked ) {
        input.checked = checked;
      } ] );
      instance.customizable = true;
    }
    if ( [ 'TText', 'TMultiLineText', 'TRichText' ].indexOf( typeName ) >= 0 ) {
      input = document.createElement( 'input' );
      input.type = 'text';
      !options.phetioReadOnly && div.appendChild( input );
      input.onkeyup = function( e ) {
        simIFrameClient.invoke( phetioID, 'setText', [ input.value ] );
      };

      var getTextButton = document.createElement( 'button' );
      getTextButton.innerHTML = 'Get';
      div.appendChild( getTextButton );

      getTextButton.addEventListener( 'click', function() {
        simIFrameClient.invoke( phetioID, 'getText', [], function( text ) {
          input.value = text;
        } );
      } );
      instance.textInput = input;
    }
    if ( getMethod( 'fire', type ) ) {

      input = document.createElement( 'button' );
      input.type = 'button';
      input.innerHTML = 'Fire';
      input.id = 'button_' + phetioID;
      div.appendChild( input );

      input.addEventListener( 'click', function() {
        simIFrameClient.invoke( phetioID, 'fire' );
      } );
    }

    return instance;
  };
})();