// Copyright 2016, University of Colorado Boulder

(function() {
  'use strict';


  // Get the initialization parameters from the json file.
  var simName = QueryStringMachine.get( 'sim', { type: 'string' } );

  // Load the phet-io wrapper script, see http://stackoverflow.com/questions/14521108/dynamically-load-js-inside-js
  var script = document.createElement( 'script' );
  script.src = 'js/' + simName + '.js';
  script.onload = function() {

    document.title = window.simConfiguration.pageTitle;

    // Sim-specific array of features in a datapoint to appear in plot and controls
    var features = window.simConfiguration.features;

    // Add the appropriate headers to the data table
    var thead = $( '.fixed_headers thead' );

    var initialTableHeaders = [ { text: 'Trial', width: 34 } ];

    // Splice sim-specific table headers into the general list
    var tableHeaders = initialTableHeaders
      .concat( window.simConfiguration.tableHeaders )
      .concat( [
        { text: 'Restore Trial' },
        { text: 'Add to Graph' },
        { text: 'Delete' } // Column for delete buttons
      ] );

    // Create a single HTML string for the row of table headers and append to DOM
    var header = '<tr>';
    tableHeaders.forEach( function( h ) {
      header += '<th>' + h.text + '</th>';
    } );
    header += '</tr>';
    thead.append( header );

    var cssFile = document.createElement( 'link' );
    cssFile.setAttribute( 'rel', 'stylesheet' );
    cssFile.setAttribute( 'type', 'text/css' );
    cssFile.setAttribute( 'href', simName + '.css' );
    document.getElementsByTagName( 'head' )[ 0 ].appendChild( cssFile );

    // Load the PhET-iO Simulation
    var sim = WrapperUtils.getSim( 'john-travoltage', '1.4.0-phetiodev.1' );
    var simFrame = document.getElementById( 'simulation-iframe' );
    var simIFrameClient = new SimIFrameClient( simFrame );
    window.simIFrameClient = simIFrameClient;

    var chartSelector = '#graph div.chart';

    /**
     * Create and configure the chart and axis controls
     * @returns {scatterPlot} - the configured scatterPlot instance
     */
    function createChart() {
      var graphDiv = d3.select( chartSelector );
      var width = graphDiv.node().getBoundingClientRect().width;
      var height = 0.618 * width; // Golden ratio

      // Create and configure the plot.
      var plot = window.scatterPlot()
        .width( width )
        .height( height )
        .margin( { top: 10, right: 25, bottom: 50, left: 85 } )
        .axisLabel( { x: '', y: '' } )
        .axisLabelOffset( { x: 40, y: 70 } )
        .axisLabelAlignment( { x: 'middle', y: 'middle' } )
        .scaleTransform( { x: 'linear', y: 'linear' } );

      // Create an empty chart (this is the state if viewed at sim startup)
      d3.select( chartSelector ).call( plot );

      // Hide the div containing the plot.  It is drawing a blank plot if initialized to hidden.
      $( '#graph .content' ).hide();

      return plot;
    }


    var selectedScaleTransform = {
      x: 'linear',
      y: 'linear'
    };

    var selectedFeatureName = {
      x: '',
      y: ''
    };

    /**
     * @param dataPoints
     * @returns a copy of dataPoints that has the selected scaleTransform applied.
     *
     * NOTE: no data validation is done here - use isDataDefinedForTransform.
     */
    function scaleData( dataPoints ) {
      return dataPoints.map( function( dataPoint ) {
        var newPoint = JSON.parse( JSON.stringify( dataPoint ) );
        switch( selectedScaleTransform.x ) {
          case 'inverse':
            newPoint[ selectedFeatureName.x ] = 1 / dataPoint[ selectedFeatureName.x ];
            break;
          case 'log' :
            newPoint[ selectedFeatureName.x ] = Math.log10( dataPoint[ selectedFeatureName.x ] );
            break;
          default:
            break;
        }

        switch( selectedScaleTransform.y ) {
          case 'inverse':
            newPoint[ selectedFeatureName.y ] = 1 / dataPoint[ selectedFeatureName.y ];
            break;
          case 'log' :
            newPoint[ selectedFeatureName.y ] = Math.log10( dataPoint[ selectedFeatureName.y ] );
            break;
          default:
            break;
        }

        return newPoint;
      } );
    }

    /**
     * Returns true if
     *  1. all selected data points exist
     *  AND
     *  2. all selected points are non-zero if the scale transform is non-linear
     *
     * @param data - {array} recordedDataPoints
     * @param transform - {Object} selectedScaleTransform
     * @returns {boolean}
     */
    function isDataDefinedForTransform( data, transform ) {
      var i;
      // Does data exist?
      for ( i = 0; i < data.length; i++ ) {
        if ( data[ i ].visible && ( (!data[ i ][ selectedFeatureName.x ] && data[ i ][ selectedFeatureName.x ] !== 0 )
                                    || (!data[ i ][ selectedFeatureName.y ] && data[ i ][ selectedFeatureName.y ] !== 0 )) ) {
          return false;
        }
      }

      if ( transform.x === 'inverse' ) {
        for ( i = 0; i < data.length; i++ ) {
          if ( data[ i ].visible && data[ i ][ selectedFeatureName.x ] === 0 ) {
            return false;
          }
        }
      }

      if ( transform.x === 'log' ) {
        for ( i = 0; i < data.length; i++ ) {
          if ( data[ i ].visible && data[ i ][ selectedFeatureName.x ] <= 0 ) {
            return false;
          }
        }
      }

      if ( transform.y === 'inverse' ) {
        for ( i = 0; i < data.length; i++ ) {
          if ( data[ i ].visible && data[ i ][ selectedFeatureName.y ] === 0 ) {
            return false;
          }
        }
      }

      if ( transform.y === 'log' ) {
        for ( i = 0; i < data.length; i++ ) {
          if ( data[ i ].visible && data[ i ][ selectedFeatureName.y ] <= 0 ) {
            return false;
          }
        }
      }

      // All checks passed, data must be valid
      return true;
    }


    /**
     * Create HTML selector elements for the features to be displayed by chart
     * and add to the page
     *
     * @param  {string} axisId - CSS selector for the HTML <select> dropdown element
     * @param  {string} axis - 'x' or 'y'
     */
    function addChartFeatureSelectors( axisId, axis ) {

      // Add a default "select x" or "select y" option
      var option = document.createElement( 'option' );
      option.setAttribute( 'disabled', 'disabled' );
      option.setAttribute( 'selected', 'selected' );
      option.innerHTML = 'Select ' + axis;
      $( axisId ).append( option );

      // Add options for the data features
      features.forEach( function( feature ) {
        option = document.createElement( 'option' );
        option.innerHTML = feature.content;
        option.setAttribute( 'value', feature.value );
        $( axisId ).append( option );
      } );
    }

    // Initialize the Recording module, which will check query parameters and prepare for data recording.
    var recorder = new window.Recorder();

    // When the recorder is ready, launch the sim.
    // TODO: for record and playback, you must have phetioEmitInputEvents
    recorder.start( function( sessionMetadata ) {

      // Launch the sim after all listeners have been wired up and after we have a session ID
      simIFrameClient.launchSim( sim.URL + '&' + window.simConfiguration.queryString, {
        expressions: window.simConfiguration.customization,
        phetioEventsListener: function( message ) {

          // Send the data
          recorder.recordMessage( message );
        },
        simStartedMetadata: sessionMetadata,
        emitInputEvents: true,
        onSimInitialized: function() {
          simIFrameClient.invoke( window.simConfiguration.simName + '.navigationBar.titleTextNode', 'setText', [ window.simConfiguration.simTitle ] );

          // When the sim is reset, re-apply the initial customization.
          simIFrameClient.invoke( window.simConfiguration.resetAllButtonFiringPropertyTandem, 'lazyLink', [ function( isFiring ) {

            // When it is no longer firing, re-apply.
            !isFiring && simIFrameClient.invokeSequence( window.simConfiguration.customization );
          } ] );
        }
      } );
    } );

    // W prefix is for wrapper
    var WButton = window.WrapperTypes.WButton;
    var WTextArea = window.WrapperTypes.WTextArea;
    var WCheckBox = window.WrapperTypes.WCheckBox;


    // Toggles visibility of rhs content when +/- is clicked
    $( '.collapsible .toggle-content' ).click( function() {
      var button = $( this );

      // Show/hide the content panel
      button.parent().find( '.content' ).toggle();

      // Get the machine readable name for the phet-io message
      var id = button.parent().attr( 'id' );
      id += ( button.attr( 'class' ).indexOf( 'plus' ) > 0 ? 'Expand' : 'Collapse' ) + 'Button';

      // Switch between + and -
      button.toggleClass( 'fa-minus-square fa-plus-square' );

      simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
        eventType: 'wrapper',
        phetioID: 'labBook.' + id,
        componentType: WButton,
        event: 'pressed'
      } ] );
    } );

    // function to add row to table containing the results of a trial
    var trialNumber = 1;

    /**
     * The given row will be swapped with the row in the previous position.
     *
     * @param tableRow - a row in '#trialTable'
     */
    function decrementRow( tableRow ) {
      var previous = tableRow.prev();
      tableRow.after( previous );
    }

    /**
     * Find the index of the object in userData
     * @param {number} tn
     * @returns {number}
     */
    function trialNumberToIndex( tn ) {
      assert && assert( typeof tn === 'number', 'Trial number should be a number, but it was a ' + ( typeof tn ) );
      var index = -1;
      for ( var i = 0;
            ( i < recordedDataPoints.length && index === -1 ); i++ ) {
        if ( recordedDataPoints[ i ].trialNumber === tn ) {
          index = i;
        }
      }
      return index;
    }

    /**
     * Creates a font-awesome object that will be recorded by phet-io via an event listener.
     *
     * @param componentType:String - the phet-io componentType of the element
     * @param trialNumber:Number - the trial id of the component
     *
     * @returns <i> element that will need a font-awesome class added by using d.className += ' fa-subclass';
     */
    function createPhetioElement( componentType, trialNumber ) {
      var button = document.createElement( 'i' );
      button.setAttribute( 'aria-hidden', 'true' );
      button.setAttribute( 'class', 'fa' );

      var buttonId = componentType + trialNumber;
      button.setAttribute( 'id', buttonId );
      button.trialNumber = trialNumber;

      return button;
    }

    function addDataPointToTable( dataPoint ) {
      var table = $( '.fixed_headers tbody' )[ 0 ];
      var row = table.insertRow( -1 );
      row.setAttribute( 'id', 'trialTableRow' + trialNumber );

      var cell0 = row.insertCell( 0 );

      var trialNumberLabel = document.createElement( 'p' );
      trialNumberLabel.innerHTML = ( trialNumber ).toString();
      var decrementButton = createPhetioElement( 'decrementButton', trialNumber );
      decrementButton.className += ' fa-arrow-circle-up';
      var incrementButton = createPhetioElement( 'incrementButton', trialNumber );
      incrementButton.className += ' fa-arrow-circle-down';

      cell0.appendChild( decrementButton );
      cell0.appendChild( trialNumberLabel );
      cell0.appendChild( incrementButton );

      window.simConfiguration.addCells( row, dataPoint );

      var screenshotCell = row.insertCell( initialTableHeaders.length + window.simConfiguration.tableHeaders.length );

      var leftPictureSpan = document.createElement( 'span' );
      leftPictureSpan.appendChild( dataPoint.image );

      var restoreButton = createPhetioElement( 'restoreButton', trialNumber );
      restoreButton.className += ' fa-play-circle';

      var rightPictureSpan = document.createElement( 'span' );
      rightPictureSpan.appendChild( restoreButton );
      var pictureDiv = document.createElement( 'div' );
      pictureDiv.appendChild( leftPictureSpan );
      pictureDiv.appendChild( rightPictureSpan );
      screenshotCell.appendChild( pictureDiv );

      var addToGraphCell = row.insertCell( 6 );
      var checkBox = document.createElement( 'input' );
      var checkBoxID = 'addToGraphCheckBox' + trialNumber;
      checkBox.setAttribute( 'id', checkBoxID );
      checkBox.setAttribute( 'type', 'checkbox' );
      checkBox.checked = false;
      checkBox.trialNumber = trialNumber;
      addToGraphCell.appendChild( checkBox );

      var deleteRowCell = row.insertCell( 7 );
      var deleteButton = createPhetioElement( 'deleteButton', trialNumber );
      deleteButton.className += ' fa-times-circle';
      deleteRowCell.appendChild( deleteButton );

      trialNumber++;

      restoreButton.addEventListener( 'click', function() {
        window.toast( 'Simulation restored to Trial ' + this.trialNumber, 4000 );

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: recordedDataPoints
          }
        } ] );
        simIFrameClient.invoke( 'phetio', 'setState', [ dataPoint.state ] );
      } );

      deleteButton.addEventListener( 'click', function() {
        // Get trialNumber from the id string
        var index = trialNumberToIndex( this.trialNumber );

        // We expected to see an object in userData with trialNumber == trialId, but something went wrong. Bail.
        window.assert( index !== -1, 'Attempted to delete trial that does not exist' );

        // Remove the trial object from the userData array and from the HTML table
        recordedDataPoints.splice( index, 1 );
        $( '#trialTableRow' + this.trialNumber ).remove();

        // Update chart to remove point
        d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: recordedDataPoints
          }
        } ] );
      } );

      decrementButton.addEventListener( 'click', function() {

        //  Swap with the previous row unless we are at the top.
        var tr = $( '#trialTableRow' + this.trialNumber );
        if ( tr.prevAll().length >= 1 ) {
          decrementRow( tr );
        }
        else {
          //Do not record decrement press if we don't decrement
          return;
        }

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: recordedDataPoints
          }
        } ] );
      } );

      incrementButton.addEventListener( 'click', function() {
        // Swap with subsequent row unless we are at the bottom
        var tr = $( '#trialTableRow' + this.trialNumber );
        if ( tr.nextAll().length >= 1 ) {
          decrementRow( tr.next() );
        }
        else {
          //Do not record increment if nothing changes
          return;
        }

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.' + this.id,
          componentType: WButton,
          event: 'pressed',
          parameters: {
            userData: recordedDataPoints
          }
        } ] );
      } );

      // Data table checkbox change listener
      $( '#' + checkBoxID ).change( function() {

        // Update visibility property of userData to match checkbox state
        var boxChecked = this.checked;
        var index = trialNumberToIndex( this.trialNumber );

        // We expected to see an object in userData with trialNumber == trialId, but something went wrong. Bail.
        window.assert( index !== -1, 'Attempted to plot/unplot trial that does not exist' );

        recordedDataPoints[ index ].visible = boxChecked;

        if ( !isDataDefinedForTransform( [ recordedDataPoints[ index ] ], selectedScaleTransform ) ) {
          window.toast( 'Cannot add this data point to the plot.  Either the data is empty, you have not selected an axis feature, or the data point is not defined for the selected scale.', null, 'OK' );
          this.checked = false;
          recordedDataPoints[ index ].visible = false;
          d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );
          return;
        }

        // Update chart since datapoint visibility has changed
        d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );

        // If this checkbox selection caused an axis range to include zero,
        // the chart will have automatically switched to a linear scale.
        // This block updates the radio buttons to match the chart.
        [ 'x', 'y' ].forEach( function( axis ) {

          // String representing radio button selection
          var choice = $( 'input[name="' + axis + '-axis-scale-transform"]:checked' ).val();

          // Switch to linear option if required
          if ( choice !== 'linear' && selectedScaleTransform[ axis ] === 'linear' ) {
            $( '#radio-' + axis + '-linear' ).prop( 'checked', true );
            var axisLabelObj = {};
            axisLabelObj[ axis ] = getAxisLabel( $( '#' + axis + '-feature option:selected' ).val() );
            plot.axisLabel( axisLabelObj );
            d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );

          }
        } );

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.' + this.id,
          componentType: WCheckBox,
          event: 'changed',
          parameters: {
            checked: boxChecked,
            userData: recordedDataPoints
          }
        } ] );
      } );
    }

    /**
     * This array is in order determined by the user, the items may be rearranged
     * with the decrement/increment buttons or removed with the delete button.
     **/
    var recordedDataPoints = [];
    var plot = createChart();

    addChartFeatureSelectors( '#x-feature', 'x' );
    addChartFeatureSelectors( '#y-feature', 'y' );

    $( '#record' ).click( function() {
      window.toast( 'An observation was recorded', 2000 );

      window.simConfiguration.recordDataPoint( simIFrameClient, function( dataPoint ) {

        dataPoint.trialNumber = trialNumber;

        recordedDataPoints.push( dataPoint );
        addDataPointToTable( dataPoint );

        // Autoscroll to bottom
        var tb = $( '.fixed_headers tbody' )[ 0 ];
        tb.scrollTop = 100 * recordedDataPoints.length;

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.recordDataButton',
          componentType: WButton,
          event: 'pressed',
          parameters: dataPoint // TODO: should screenshot be omitted from the data stream for performance reasons?
        } ] );
      } );
    } );

    // Notes should appear in the data stream
    var previousNotesText = '';

    $( '#notes textarea' ).on( 'change keyup paste', function() {
      var currentNotesText = $( this ).val();

      // Only send messages when the notes change
      if ( currentNotesText !== previousNotesText ) {

        simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
          eventType: 'wrapper',
          phetioID: 'labBook.textArea',
          componentType: WTextArea,
          event: 'changed',
          parameters: {
            text: currentNotesText
          }
        } ] );
        previousNotesText = currentNotesText;
      }
    } );

    // Radio button listeners for axis scale transform selection (linear, inverse, log)
    [ 'x', 'y' ].forEach( function( axis ) {
        $( '#graph input[name="' + axis + '-axis-scale-transform"]' ).change( function() {

          var axisLabelObj = {};

          if ( $( '#graph select#' + axis + '-feature' ).val() !== null ) {

            // String representing radio button selection
            var choice = $( 'input[name="' + axis + '-axis-scale-transform"]:checked' ).val();
            // Before the radio buttons have been selected, choice will be undefined.
            choice = (choice === undefined) ? 5 : choice;
            selectedScaleTransform[ axis ] = choice;

            if ( isDataDefinedForTransform( recordedDataPoints, selectedScaleTransform ) ) {
              // Append a scale type to the axis as it changes in the radio buttons
              var axisAddition = ( choice === 'inverse' ) ? '1/' : ( choice === 'log' ) ? 'log(' : '';
              var baseAxisLabel = getAxisLabel( $( '#' + axis + '-feature option:selected' ).val() );
              var axisLabel = axisAddition + baseAxisLabel;
              axisLabelObj[ axis ] = (axisAddition.includes( 'log' )) ? axisLabel + ')' : axisLabel;
              plot.axisLabel( axisLabelObj );
            }
            else {
              window.toast( 'Cannot use this scale because a data point(s) is undefined. This happens when you divide by zero or take log(0)', 0, 'OK' );
              selectedScaleTransform[ axis ] = 'linear';

              setTimeout( function() { $( '#radio-' + axis + '-linear' ).prop( 'checked', true ); }, 200 );

              // Change the axis label back to the linear version
              axisLabelObj[ axis ] = baseAxisLabel;
              plot.axisLabel( axisLabelObj );
            }
            d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );

            simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
              eventType: 'wrapper',
              phetioID: 'labBook.' + axis + 'Transform',
              componentType: WTextArea,
              event: 'changed',
              parameters: {
                feature: selectedScaleTransform[ axis ]
              }
            } ] );
          }
        } );
      }
    );

    // Dig the `content` property out of the features array from the `value` property
    function getAxisLabel( featureName ) {
      var content = '';
      features.forEach( function( feature ) {
        if ( feature.value === featureName ) {
          content = feature.content;
        }
      } );
      return content;
    }

    // Data feature selection menu callback (x-axis)
    $( '#x-feature' ).change( function() {
      selectedFeatureName.x = $( '#x-feature option:selected' ).val();
      selectedScaleTransform.x = 'linear';

      // When variable/feature changes, the chart switches to a linear scale. Keep radio button consistent.
      $( '#radio-x-linear' ).prop( 'checked', true );

      plot
        .featureName( { x: selectedFeatureName.x } )
        .scaleTransform( { x: 'linear' } )
        .axisLabel( { x: getAxisLabel( selectedFeatureName.x ) } );

      d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );

      simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
        eventType: 'wrapper',
        phetioID: 'labBook.xFeature',
        componentType: WTextArea,
        event: 'changed',
        parameters: {
          feature: selectedFeatureName.x
        }
      } ] );
    } );

    // Data feature selection menu callback (y-axis)
    $( '#y-feature' ).change( function() {
      selectedFeatureName.y = $( '#y-feature option:selected' ).val();
      selectedScaleTransform.y = 'linear';

      // When variable/feature changes, the chart switches to a linear scale. Keep radio button consistent.
      $( '#radio-y-linear' ).prop( 'checked', true );

      plot
        .featureName( { y: selectedFeatureName.y } )
        .scaleTransform( { y: 'linear' } )
        .axisLabel( { y: getAxisLabel( selectedFeatureName.y ) } );

      d3.select( chartSelector ).datum( scaleData( recordedDataPoints ) ).call( plot );

      simIFrameClient.invoke( 'phetio', 'triggerEvent', [ {
        eventType: 'wrapper',
        phetioID: 'labBook.yFeature',
        componentType: WTextArea,
        event: 'changed',
        parameters: {
          feature: selectedFeatureName.y
        }
      } ] );
    } );

  };

  document.head.appendChild( script );

})();